(*----------------------------------------------------------------------------
	Author: Edouard Hibon
	Updated on: November , 1994.
------------------------------------------------------------------------------*)

specification Cnd_Un [PotsGate,PotsTalk,PotsHang]

:noexit

library Set, Boolean, DecDigit, NaturalNumber, BasicNaturalNumber endlib

type UsersList is DecDigit, Boolean, BasicNaturalNumber
sorts List

   opns {} : -> List
	InsertL, RemoveL                        : DecDigit, List -> List
	_IsIn_, _NotIn_                         : DecDigit, List -> Bool
	_Union_, _Ints_, _Minus_                : List, List     -> List
	_eq_, _ne_, _Includes_, _IsSubsetOf_    : List, List     -> Bool
	Card                                    : List           -> Nat
	First                                   : List           -> DecDigit

   eqns  forall x, y : DecDigit,
	      s, t : List

   ofsort List
	x IsIn InsertL(y,s) =>
	InsertL(x, InsertL(y,s)) = InsertL(y,s) ;
	RemoveL(x, {}) = {} ;
	RemoveL(x, InsertL(x,s)) = s ;
	x ne y =>
	RemoveL(x,InsertL(y,s)) = InsertL(y,RemoveL(x,s)) ;
	{} Union s = s ;
	InsertL(x,s) Union t = InsertL(x,s Union t) ;
	{} Ints s = {} ;
	x IsIn t =>
	InsertL(x,s) Ints t = InsertL(x,s Ints t) ;
	x NotIn t =>
	InsertL(x,s) Ints t = s Ints t ;
	s Minus {} = s ;
	s Minus InsertL(x, t) = RemoveL(x,s) Minus t ;

   ofsort Bool
	x IsIn {} = false ;
	x eq y =>
	x IsIn InsertL(y,s) = true ;
	x ne y =>
	x IsIn InsertL(y,s) = x IsIn s ;
	x NotIn s = not(x IsIn s) ;
	s Includes {} = true ;
	s Includes InsertL(x,t) = (x IsIn s) and (s Includes t) ;
	s IsSubsetOf t = t Includes s ;
	s eq t = (s Includes t) and (t Includes s);
	s ne t = not(s eq t) ;

   ofsort DecDigit
	First({}) = 0 ;
	x NotIn s =>
	First(InsertL(x,s)) = x ;
	ofsort Nat
	Card ({}) = 0 ;
	x NotIn s =>
	Card (InsertL(x,s)) = Succ(Card(s)) ;
	endtype

type EventSeq is NaturalNumber, Boolean
sorts EventSeq

   opns 0, 1, 2, 3, 4, 5, 6, 7, 8, 9       :                    -> EventSeq
	_eq_, _ne_, _lt_, _le_, _ge_, _gt_ : EventSeq, EventSeq -> Bool
	NatNum                             : EventSeq           -> Nat

   eqns forall x, y, z : EventSeq

   ofsort Nat
	NatNum(0) = 0               ;
	NatNum(1) = Succ(NatNum(0)) ;
	NatNum(2) = Succ(NatNum(1)) ;
	NatNum(3) = Succ(NatNum(2)) ;
	NatNum(4) = Succ(NatNum(3)) ;
	NatNum(5) = Succ(NatNum(4)) ;
	NatNum(6) = Succ(NatNum(5)) ;
	NatNum(7) = Succ(NatNum(6)) ;
	NatNum(8) = Succ(NatNum(7)) ;
	NatNum(9) = Succ(NatNum(8)) ;

   ofsort  Bool
	x eq y  = NatNum(x) eq NatNum(y) ;
	x ne y  = NatNum(x) ne NatNum(y) ;
	x lt y  = NatNum(x) lt NatNum(y) ;
	x le y  = NatNum(x) le NatNum(y) ;
	x ge y  = NatNum(x) ge NatNum(y) ;
	x ge y  = NatNum(x) ge NatNum(y) ;

endtype

type DigitPair is DecDigit, Boolean
sorts DigitPair

   opns
	None    : -> DecDigit
	Pair:DecDigit,DecDigit -> DigitPair
	_eq_: DigitPair, DigitPair -> Bool

   eqns forall f1,f2,l1,l2:DecDigit , f3:Nat

   ofsort Bool
	f1 eq None     = false;
	None     eq f1 = false;
	Pair(f1,l1) eq Pair(f2,l2) = (f1 eq f2) and (l1 eq l2);

   ofsort Nat
	NatNum(None) =  0
	endtype

type SetOfPairs  is DigitPair, NaturalNumber
sorts SetOfPairs

   opns
	{}:                                          -> SetOfPairs
	InsertPair:        DigitPair,SetOfPairs  -> SetOfPairs
	RemovePair:        DigitPair,SetOfPairs  -> SetOfPairs
	InsertPairAtTheEnd:DigitPair,SetOfPairs  -> SetOfPairs
	IsPairIn :         DigitPair,SetOfPairs  -> Bool
	SecondElement:     DecDigit,SetOfPairs   -> DecDigit
	(* given the first element, find the second element *)
	FirstElement:      DecDigit,SetOfPairs   -> DecDigit
	(* given the secondelement, find the first element *)
	IsPairNotIn:       DigitPair,SetOfPairs  -> Bool
	CardPair:          DecDigit,SetOfPairs   -> Nat
	(* Number of pairs with the same first element Insert the set *)

   eqns forall x,y: DigitPair,
	s: SetOfPairs,
	f1, f2, s1, s2: DecDigit

   ofsort SetOfPairs
	InsertPairAtTheEnd(x,{})              = InsertPair(x,{});
	InsertPairAtTheEnd(x,InsertPair(x,s)) = InsertPair(x,s) ;
	not(x eq y) =>
	InsertPairAtTheEnd(x,InsertPair(y,s)) =
	InsertPair(y,InsertPairAtTheEnd(x,s));

	RemovePair(x,{}) = {};
	RemovePair(x,InsertPair(x,s)) = s;
	not(x eq y) =>
	RemovePair(x,InsertPair(y,s)) = InsertPair(y,RemovePair(x,s));

   ofsort Bool
	IsPairIn(x,{}) = false;
	IsPairIn(x,InsertPair(y,s)) = (x eq y) or (IsPairIn(x,s));
	IsPairIn(x,RemovePair(x,s)) = false;
	not(x eq y) =>
	IsPairIn(x,RemovePair(y,s)) = IsPairIn(x,s);
	IsPairNotIn(x,s) = not(IsPairIn(x,s));

   ofsort DecDigit
	SecondElement (f1,{}) = None     ;
	SecondElement (f1,InsertPair(Pair(f1,s1),s)) = s1;
	f1 ne f2 =>
	SecondElement (f1,InsertPair(Pair(f2,s2),s)) = SecondElement (f1,s);
	FirstElement (f1,{}) = None     ;     (* not tested *)
	FirstElement (s1,InsertPair(Pair(f1,s1),s)) = f1;
	s1 ne s2 =>
	FirstElement (s1,InsertPair(Pair(f2,s2),s)) = FirstElement (s1,s);

   ofsort  Nat
	CardPair (f1, {})                = 0;
	IsPairNotIn(Pair(f2, s1), s) and (f1 eq f2) =>
	CardPair (f1, InsertPair(Pair(f2, s1), s)) = Succ(CardPair(f1,s));
	IsPairNotIn(Pair(f2, s1), s) and not(f1 eq f2) =>
	CardPair (f1, InsertPair(Pair(f2, s1), s)) = CardPair(f1,s);
	endtype

type DecSet0 is Set
   actualizedby DecDigit
   using sortnames  DecDigit for Element
	              Bool for FBool
endtype

type DecSet is DecSet0
   renamedby sortnames  DecSet for Set
endtype

type Signals is Boolean, NaturalNumber
   sorts  Signal
   opns  OffHookToCall,
	 OffHookToAnswer,
	 HangsUp,
	 TalksTo,
	 RingsFrom,
	 RingsBackFrom,
	 StopToRingFrom,
	 Dials,
	 GetsTone,
	 GetsBusySignalFrom,
	 Displayed,
	 NotDisplayed    :-> Signal

	 _IsEq_          : Signal, Signal -> Bool
	 _IsNe_          : Signal, Signal -> Bool
	 Map             : Signal -> Nat
  
   eqns     forall x, y: Signal

   ofsort  Nat 
	 Map (OffHookToCall)  =  0;
	 Map (OffHookToAnswer)=  Succ(Map (OffHookToCall));
	 Map (HangsUp)        =  Succ(Map (OffHookToAnswer));
	 Map (TalksTo)        =  Succ(Map (HangsUp));
	 Map (RingsFrom)      =  Succ(Map (TalksTo));
	 Map (RingsBackFrom)  =  Succ(Map (RingsFrom));
	 Map (StopToRingFrom) =  Succ(Map (RingsBackFrom));
	 Map (Dials)          =  Succ(Map (StopToRingFrom));
	 Map (GetsTone)       =  Succ(Map (Dials));
	 Map (GetsBusySignalFrom) =  Succ(Map (GetsTone));
	 Map (Displayed)      =  Succ(Map (GetsBusySignalFrom));
	 Map (NotDisplayed)   =  Succ(Map (Displayed));

   ofsort  Bool
	 x IsEq y = Map (x) eq Map (y);
	 x IsNe y = Map (x) ne Map (y);
endtype

behaviour 

(
	Cnd_Un [PotsGate,PotsTalk,PotsHang] 
)

where

     process Cnd_Un [PotsGate,PotsTalk,PotsHang]
     :noexit:=
     (
        InitConnection [PotsGate,PotsTalk,PotsHang] 
        ||
        NewGlobalView [PotsGate,PotsTalk,PotsHang] 
     )
     endproc (* Cnd_Un *)

     process InitConnection [PotsGate,PotsTalk,PotsHang] 
     :noexit:=
     (
	InitNewConnection [PotsGate,PotsTalk,PotsHang]
	|||
	i; InitConnection [PotsGate,PotsTalk,PotsHang] 
     )
     endproc (* InitConnection *)

     process InitNewConnection [PotsGate,PotsTalk,PotsHang]
     :noexit:=
     (
        PotsGate ?Caller:DecDigit !OffHookToCall ;
	CallerGetsTone [PotsGate,PotsTalk,PotsHang] (Caller)
	>> stop 
     )
     endproc (* InitNewConnection *)

     process CallerGetsTone [PotsGate,PotsTalk,PotsHang]
			    (Caller:DecDigit)
     :exit:=
     (        
        PotsGate  !Caller  !GetsTone ;
        (
           PotsGate  !Caller  !Dials  ?Called:DecDigit ;
           (
	      CalledRingsFromCaller [PotsGate,PotsTalk,PotsHang]
                                    (Caller,Called)
	   []
	      CallerGetsBusySignalFromCalled [PotsGate,PotsHang]
                                             (Caller,Called)
	   []
              UserHangsUp [PotsHang] (Caller)
           )
           [] UserHangsUp [PotsHang] (Caller)
        )
        [] UserHangsUp [PotsHang] (Caller)
     )
     endproc (* InitNewConnection *)

     process UserHangsUp [Gate] (User:DecDigit)
     :exit:=
     (
        Gate  !User  !HangsUp ;
        exit
     )
     endproc (* UserHangsUp *)

     process CallerGetsBusySignalFromCalled [PotsGate,PotsHang]
                                            (Caller,Called:DecDigit)
     :exit:=
     (
        PotsGate  !Caller !GetsBusySignalFrom !Called ;
        UserHangsUp [PotsHang] (Caller)
     )
     endproc (* CallerGetsBusySignalFromCalled *)

     process CalledRingsFromCaller [PotsGate,PotsTalk,PotsHang]
                                   (Caller,Called:DecDigit)
     :exit:=
     (
        PotsGate  !Called  !RingsFrom  !Caller  ?signal:Signal ;
        (
           PotsGate  !Caller  !RingsBackFrom  !Called ;
           (
                PotsGate  !Called  !OffHookToAnswer  !Caller ;
                CallerTalksToCalled [PotsTalk,PotsHang]
                                    (Caller,Called)
             [] UserHangsUp [PotsHang] (Caller)
           )
           [] UserHangsUp [PotsHang] (Caller)
        )
     )
     endproc (* CalledRingsFromCaller *)

     process CallerTalksToCalled [PotsTalk,PotsHang]
				  (Caller,Called:DecDigit)
     :exit:= 
     (
	Conversation [PotsTalk] (Caller,Called) 
	[>
	(
	   (
	      PotsHang  !Caller  !HangsUp ;
	      UserHangsUp [PotsHang] (Called) 
	   )
          []
	   (
	      PotsHang  !Called  !HangsUp ;
	      UserHangsUp [PotsHang] (Caller) 
	   )
	)
     )
     endproc (* CallerTalksToCalled *)                 

     process Conversation [Gate] (Caller,Called:DecDigit)
     :noexit:=
     (
        Gate  !Caller   !TalksTo  !Called ;
        Conversation [Gate] (Caller,Called)
     []
        Gate  !Called   !TalksTo  !Caller ;
        Conversation [Gate] (Caller,Called)
     )
     endproc (* Conversation *)


     process NewGlobalView [PotsGate,PotsTalk,PotsHang]
     :noexit:=
     (
	GlobalView [PotsGate,PotsTalk,PotsHang]
            (
                {} of DecSet ,
                {} of SetOfPairs ,
		insert (2, {} of Decset),
                Insert (4, {} of Decset)
            )
     )
     endproc (* NewGlobalView *)
	   
     process GlobalView [PotsGate,PotsTalk,PotsHang]
	     (
	       BusyList   : DecSet,	(* list of users on line   	  *)
	       WaitPairs  : SetOfPairs,	(* list of pairs (X,Y) where Y is *)
					(* waiting for X to answer	  *)
	       CndList ,  		(* list of users with Calling_Num *)
					(* ber_Delivery feature		  *)
               UnList    : DecSet       (* list of users with Unlisted_   *)
                                        (* Number feature           	  *)
	     )
     :noexit:=
     (
	Phases [PotsGate,PotsTalk,PotsHang]
         (
	 BusyList,
	 WaitPairs,
	 CndList,
	 UnList
         )
     [] PrintSets [PotsGate,PotsTalk,PotsHang]
        (
	 BusyList,
	 WaitPairs,
         CndList,
         UnList
        )
     )
     endproc (* GlobalView *)

     process Phases[PotsGate,PotsTalk,PotsHang]
		   (
		     BusyList   : DecSet,
	 	     WaitPairs  : SetOfPairs,
		     CndList ,
         	     UnList     : DecSet
		   )
     :noexit:=
     (
	PotsGate  ?Caller:DecDigit  !OffHookToCall 
	[     (Caller NotIn BusyList)
	  and (SecondElement(Caller,WaitPairs) eq none)
	] ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
		   (
		     Insert (Caller,BusyList),
		     InsertPairAtTheEnd (Pair (Caller,Caller),WaitPairs),
		     CndList,
         	     UnList
		   )
    []  PotsGate  ?Caller:DecDigit  !GetsTone ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
	           (
		     BusyList,
		     WaitPairs,
                     CndList,
                     UnList
	           )
    []  PotsGate  ?Caller:DecDigit  !Dials  ?Called:DecDigit ; 
	GlobalView [PotsGate,PotsTalk,PotsHang]
		   (
		     BusyList,
		     WaitPairs,
                     CndList,
                     UnList
		   )
    []  PotsGate ?Caller:DecDigit !GetsBusySignalFrom ?X:DecDigit
        [    (X IsIn BusyList)
          or (      (SecondElement(X,WaitPairs) ne none)
                and (SecondElement(X,WaitPairs) ne Caller)
             )
        ] ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
                   (
                     BusyList,
                     WaitPairs,
                     CndList,
                     UnList
                   )
    []  PotsGate  ?Called:DecDigit !RingsFrom ?Caller:DecDigit !Displayed 
        [     (Called NotIn BusyList)
          and (    (SecondElement(Called,WaitPairs) eq none)
                or (SecondElement(Called,WaitPairs) eq Caller)
              )
	  and (Called IsIn CndList)
	] ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
		   (
		     BusyList,
		     InsertPairAtTheEnd (Pair (Called,Caller) , 
                      RemovePair ( Pair (Caller,Caller),WaitPairs)),
                     CndList,
                     UnList
		   )
    []  PotsGate  ?Called:DecDigit !RingsFrom ?Caller:DecDigit !NotDisplayed
        [     (Called NotIn BusyList)
          and (    (SecondElement(Called,WaitPairs) eq none)
                or (SecondElement(Called,WaitPairs) eq Caller)
              )
          and (    (Called NotIn CndList)
		or 
                   (Caller IsIn UnList)
	      )
        ] ;
        GlobalView [PotsGate,PotsTalk,PotsHang]
                   (
                     BusyList,
                     InsertPairAtTheEnd (Pair (Called,Caller) ,
                      RemovePair ( Pair (Caller,Caller),WaitPairs)),
                     CndList,
                     UnList
                   )
    []  PotsGate  ?Caller:DecDigit  !RingsBackFrom  ?Called:DecDigit ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
                   (
                     BusyList,
                     WaitPairs,
                     CndList,
                     UnList
                   )
    []  PotsGate  ?Called:DecDigit  !OffHookToAnswer  ?Caller:DecDigit ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
		   (
	 	     Insert (Called,BusyList),
	 	     RemovePair (Pair (Called,Caller) , WaitPairs),
                     CndList,
                     UnList
		   )
    []  PotsGate  ?Called :DecDigit  !StopToRingFrom  ?Caller :DecDigit
        [Caller NotIn BusyList] ;
        GlobalView [PotsGate,PotsTalk,PotsHang] 
                   ( 
                     BusyList,
                     RemovePair (Pair (Called,Caller) , WaitPairs ) ,
                     CndList,
                     UnList
                   )
    []  PotsHang  ?user:DecDigit  !HangsUp ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
		   ( 
	 	     Remove (user,BusyList),
	 	     RemovePair (Pair (FirstElement (user,WaitPairs),user),
		      WaitPairs),
                     CndList,
                     UnList
		   )
    []  PotsTalk  ?user1:DecDigit  !TalksTo  ?user2:DecDigit ;
	GlobalView [PotsGate,PotsTalk,PotsHang]
                   (
                     BusyList,
                     WaitPairs,
                     CndList,
                     UnList
                   )
     )
     endproc (* Phases *)


     process PrintSets
	[PotsGate,PotsTalk,PotsHang]
	(BusyL : DecSet, WaitP : SetOfPairs, CndL, UnL : DecSet)
     :noexit:=
     (
	hide BusyList,WaitPairs,CndList,UnList in
	(
             BusyList   !BusyL ;
	     GlobalView [PotsGate,PotsTalk,PotsHang]
         		( BusyL, WaitP, CndL, UnL )
         []  WaitPairs  !WaitP ;
	     GlobalView [PotsGate,PotsTalk,PotsHang]
         		( BusyL, WaitP, CndL, UnL )
         []  CndList    !CndL ;
	     GlobalView [PotsGate,PotsTalk,PotsHang]
         		( BusyL, WaitP, CndL, UnL )
         []  UnList    !UnL ;
             GlobalView [PotsGate,PotsTalk,PotsHang]
                        ( BusyL, WaitP, CndL, UnL )
	)
     )
     endproc (* PrintSets *)

endspec

