(*----------------------------------------------------------------------------
Author:   Edouard Hibon
Updated on:  November , 1994.
------------------------------------------------------------------------------*)

specification Cw_Cfbl [PotsGate,PotsTalk,PotsHang,CwLink]

:noexit

library Set, Boolean, DecDigit, NaturalNumber, BasicNaturalNumber endlib

type    UsersList is DecDigit, Boolean, BasicNaturalNumber

sorts   List

opns    {}                                      : -> List
	InsertL, RemoveL                        : DecDigit, List -> List
	_IsIn_, _NotIn_                         : DecDigit, List -> Bool
	_Union_, _Ints_, _Minus_                : List, List -> List
	_eq_, _ne_, _Includes_, _IsSubsetOf_    : List, List -> Bool
	Card                                    : List -> Nat
	First                                   : List -> DecDigit

eqns    forall  x, y    : DecDigit,
	        s, t    : List

	ofsort  List

	        x IsIn InsertL(y,s) =>
	InsertL(x, InsertL(y,s))  = InsertL(y,s)                   ;

	RemoveL(x, {}) = {}                                      ;
	RemoveL(x, InsertL(x,s))  = s                             ;
	        x ne y =>
	RemoveL(x, InsertL(y,s))  = InsertL(y, RemoveL(x,s))        ;

	{} Union s              = s                             ;
	InsertL(x,s) Union t     = InsertL(x,s Union t)           ;

	{} Ints s               = {}                            ;
	        x IsIn t =>
	InsertL(x,s) Ints t      = InsertL(x,s Ints t)            ;
	        x NotIn t =>
	InsertL(x,s) Ints t      = s Ints t                      ;

	s Minus {}              = s                             ;
	s Minus InsertL(x, t)    = RemoveL(x,s) Minus t           ;

	ofsort  Bool
	x IsIn {}               = false                         ;
	        x eq y =>
	x IsIn InsertL(y,s)      = true                          ;
	        x ne y =>
	x IsIn InsertL(y,s)      = x IsIn s                      ;

	x NotIn s               = not(x IsIn s)                 ;

	s Includes {}           = true                          ;
	s Includes InsertL(x,t)  = (x IsIn s) and (s Includes t) ;

	s IsSubsetOf t          = t Includes s                  ;

	s eq t                  = (s Includes t) and (t Includes s);

	s ne t                  = not(s eq t)                   ;

	ofsort  DecDigit

	First({})                = 0                         ;
	        x NotIn s =>
	First(InsertL(x,s))       = x                            ;

	ofsort  Nat
	Card ({})                = 0                            ;
	        x NotIn s =>
	Card (InsertL(x,s))       = Succ(Card(s))                      ;
endtype

type    EventSeq is NaturalNumber, Boolean
sorts   EventSeq
opns    0, 1, 2, 3, 4, 5, 6, 7, 8, 9            : -> EventSeq
	_eq_, _ne_, _lt_, _le_, _ge_, _gt_      : EventSeq, EventSeq -> Bool
	NatNum                                  : EventSeq -> Nat

eqns    forall  x, y, z : EventSeq

	ofsort  Nat

	NatNum(0)       = 0                     ;
	NatNum(1)       = Succ(NatNum(0))       ;
	NatNum(2)       = Succ(NatNum(1))       ;
	NatNum(3)       = Succ(NatNum(2))       ;
	NatNum(4)       = Succ(NatNum(3))       ;
	NatNum(5)       = Succ(NatNum(4))       ;
	NatNum(6)       = Succ(NatNum(5))       ;
	NatNum(7)       = Succ(NatNum(6))       ;
	NatNum(8)       = Succ(NatNum(7))       ;
	NatNum(9)       = Succ(NatNum(8))       ;

	ofsort  Bool

	x eq y  = NatNum(x) eq NatNum(y)        ;
	x ne y  = NatNum(x) ne NatNum(y)        ;
	x lt y  = NatNum(x) lt NatNum(y)        ;
	x le y  = NatNum(x) le NatNum(y)        ;
	x ge y  = NatNum(x) ge NatNum(y)        ;
	x ge y  = NatNum(x) ge NatNum(y)        ;

endtype

type DigitPair is DecDigit, Boolean

   sorts DigitPair
   opns
       None    : -> DecDigit
       Pair:DecDigit,DecDigit -> DigitPair
       _eq_: DigitPair, DigitPair -> Bool

   eqns

       forall f1,f2,l1,l2:DecDigit , f3:Nat

       ofsort Bool
	  f1 eq None     = false;
	  None     eq f1 = false;
	  Pair(f1,l1) eq Pair(f2,l2) = (f1 eq f2) and (l1 eq l2);

       ofsort Nat
	  NatNum(None    ) =  0
endtype

type SetOfPairs  is DigitPair, NaturalNumber
     sorts SetOfPairs
     opns
	{}:                                          -> SetOfPairs
	InsertPair:            DigitPair,SetOfPairs  -> SetOfPairs
	RemovePair:            DigitPair,SetOfPairs  -> SetOfPairs
	InsertPairAtTheEnd:    DigitPair,SetOfPairs  -> SetOfPairs
	IsPairIn :             DigitPair,SetOfPairs  -> Bool
	SecondElement:         DecDigit,SetOfPairs   -> DecDigit
	(* given the first element, find the second element *)
	FirstElement:          DecDigit,SetOfPairs   -> DecDigit
	(* given the second element, find the first element *)
	IsPairNotIn:           DigitPair,SetOfPairs  -> Bool
	CardPair:              DecDigit,SetOfPairs   -> Nat
	(* Number of pairs with the same first element Insert the set *)

   eqns forall x,y: DigitPair,
	         s: SetOfPairs,
    f1, f2, s1, s2: DecDigit

   ofsort SetOfPairs
	InsertPairAtTheEnd(x,{})              = InsertPair(x,{});
	InsertPairAtTheEnd(x,InsertPair(x,s)) = InsertPair(x,s) ;
	not(x eq y) =>
	InsertPairAtTheEnd(x,InsertPair(y,s)) =
	                                 InsertPair(y,InsertPairAtTheEnd(x,s));

	RemovePair(x,{}) = {};
	RemovePair(x,InsertPair(x,s)) = s;
	not(x eq y) =>
	RemovePair(x,InsertPair(y,s)) = InsertPair(y,RemovePair(x,s));

   ofsort Bool
	IsPairIn(x,{}) = false;
	IsPairIn(x,InsertPair(y,s)) = (x eq y) or (IsPairIn(x,s));
	IsPairIn(x,RemovePair(x,s)) = false;
	not(x eq y) =>
	IsPairIn(x,RemovePair(y,s)) = IsPairIn(x,s);
	IsPairNotIn(x,s) = not(IsPairIn(x,s));

   ofsort DecDigit
	SecondElement (f1,{}) = None     ;
	SecondElement (f1,InsertPair(Pair(f1,s1),s)) = s1;
	f1 ne f2 =>
	SecondElement (f1,InsertPair(Pair(f2,s2),s)) = SecondElement (f1,s);

	FirstElement (f1,{}) = None     ;     (* not tested *)
	FirstElement (s1,InsertPair(Pair(f1,s1),s)) = f1;
	s1 ne s2 =>
	FirstElement (s1,InsertPair(Pair(f2,s2),s)) = FirstElement (s1,s);

   ofsort  Nat
	CardPair (f1, {})                = 0;
	IsPairNotIn(Pair(f2, s1), s) and (f1 eq f2) =>
	CardPair (f1, InsertPair(Pair(f2, s1), s)) = Succ(CardPair(f1,s));
	IsPairNotIn(Pair(f2, s1), s) and not(f1 eq f2) =>
	CardPair (f1, InsertPair(Pair(f2, s1), s)) = CardPair(f1,s);
endtype

type DecSet0 is Set
     actualizedby DecDigit
     using
     sortnames DecDigit for Element
	       Bool for FBool
endtype

type DecSet is DecSet0
     renamedby
     sortnames  DecSet for Set
endtype

type Signal is Boolean, NaturalNumber
     sorts  Signal
     opns OffHookToCall,
	  OffHookToAnswer,
	  HangsUp,
	  TalksTo,
	  RingsFrom,
	  StopToRingFrom,
	  RingsBackFrom,
	  Dials,
	  GetsTone,
	  GetsBusySignalFrom,
	  GetsCwTone,
	  FlashHook,
	  NotOnLine,
	  OnLine,
	  InsteadOf       :-> Signal

	  _IsEq_          : Signal, Signal -> Bool
	  _IsNe_          : Signal, Signal -> Bool
	  Map             : Signal -> Nat

     eqns forall x, y: Signal

     ofsort  Nat
	Map (OffHookToCall)       =  0;
	Map (OffHookToAnswer)     =  Succ(Map (OffHookToCall));
	Map (HangsUp)             =  Succ(Map (OffHookToAnswer));
	Map (TalksTo)             =  Succ(Map (HangsUp));
	Map (RingsFrom)           =  Succ(Map (TalksTo));
	Map (StopToRingFrom)      =  Succ(Map (RingsFrom));
	Map (RingsBackFrom)       =  Succ(Map (StopToRingFrom));
	Map (Dials)               =  Succ(Map (RingsBackFrom));
	Map (GetsTone)            =  Succ(Map (Dials));
	Map (GetsBusySignalFrom)  =  Succ(Map (GetsTone));
	Map (GetsCwTone)          =  Succ(Map (GetsBusySignalFrom));
	Map (FlashHook)           =  Succ(Map (GetsCwTone));
	Map (NotOnLine)           =  Succ(Map (FlashHook));
	Map (OnLine)              =  Succ(Map (NotOnLine));
	Map (InsteadOf)           =  Succ(Map (OnLine));

    ofsort  Bool
	x IsEq y =  Map   (x) eq Map   (y);
	x IsNe y =  Map   (x) ne Map   (y);
endtype


behaviour

(
     Cw_Cfbl [PotsGate,PotsTalk,PotsHang,CwLink]
)

where

     process Cw_Cfbl [PotsGate,PotsTalk,PotsHang,CwLink]
     :noexit:=
     (
	hide HiddenGate in
	(
               InitConnection [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
               ||
               NewGlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
	)
     )
     endproc (* Cw_Cfbl *)

     process InitConnection [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
     :noexit:=
     (
          (
             InitNewConnection [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
             >> InitConnection [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
          )
          |||
          i; InitConnection [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
     )
     endproc (* InitConnection *)


     process InitNewConnection [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
     :exit:=
     (
        PotsGate  ?Caller : DecDigit   !OffHookToCall ;
        (
           PotsGate  !Caller  !GetsTone ;
           (
              PotsGate  !Caller  !Dials  ?Called:DecDigit ;
	      HiddenGate  ?X:DecDigit  !Called ;
              (
                  CalledRingsFromCaller [PotsGate,PotsTalk,PotsHang,CwLink]
                                        (Caller,Called,Called)
               []
		  [X ne none] ->
                  CalledRingsFromCaller [PotsGate,PotsTalk,PotsHang,CwLink]
                                        (Caller,Called,X)
               []
                  CalledGetsCwToneFromCaller [PotsGate,PotsTalk,PotsHang,CwLink]
                                             (Caller,Called)
               []
                  CallerGetsBusySignalFromCalled [PotsGate,PotsHang]
                                                 (Caller,Called,Called)
               []
		  [X ne none] ->
                  CallerGetsBusySignalFromCalled [PotsGate,PotsHang]
                                                 (Caller,Called,X)
               []
                  UserHangsUpOnPotsHang [PotsHang] (Caller)
              )
              []  UserHangsUpOnPotsHang [PotsHang] (Caller)
           )
           []  UserHangsUpOnPotsHang [PotsHang] (Caller)
        )
     )
     endproc (* InitNewConnection *)

     process UserHangsUpOnPotsHang [PotsHang] (User:DecDigit)
     :exit:=
     (
        PotsHang  !User  !HangsUp ;
        exit
     )
     endproc (* UserHangsUpOnPotsHang *)

     process CallerGetsBusySignalFromCalled [PotsGate,PotsHang]
                                            (Caller,Called,X:DecDigit)
     :exit:=
     (
        PotsGate !Caller !GetsBusySignalFrom !X !InsteadOf !Called ;
        UserHangsUpOnPotsHang [PotsHang] (Caller)
     )
     endproc (* CallerGetsBusySignalFromCalled *)

     process CalledRingsFromCaller [PotsGate,PotsTalk,PotsHang,CwLink]
                                   (Caller,Called,X:DecDigit)
     :exit:=
     (
        PotsGate !X !InsteadOf !Called !RingsFrom !Caller ;
        (
           PotsGate  !Caller  !RingsBackFrom  !X ;
           (
		PotsGate  !X  !OffHookToAnswer  !Caller ;
		CallerTalksToCalled [PotsGate,PotsTalk,PotsHang,CwLink]
				    (Caller,X)
            []  UserHangsUpOnPotsHang [PotsHang] (Caller)
           )
          []    UserHangsUpOnPotsHang [PotsHang] (Caller)
        )
     )
     endproc (* CalledRingsFromCaller *)

     process CallerTalksToCalled [PotsGate,PotsTalk,PotsHang,CwLink]	(* B *)
				 (Caller,Called : DecDigit)
     :exit:=
     (
        PotsConversation [PotsTalk] (Caller,Called)
        [>
        (
           (
              PotsHang  !Caller  !HangsUp ;
              (
		  UserHangsUpOnPotsHang [PotsHang] (Called)
	       []
		  CwLink  !Called  !GetsCwTone;
		  XGetsCwToneAndYHangsUp [PotsHang,CwLink] (Called,Caller)
	      )
           )
          []
           (
              PotsHang  !Called  !HangsUp ;
              (
		  UserHangsUpOnPotsHang [PotsHang] (Caller)
	       []
		  CwLink  !Caller  !GetsCwTone ;
		  XGetsCwToneAndYHangsUp [PotsHang,CwLink] (Caller,Called)
	      )
           )
          []
           (
              CwLink  !Caller  !GetsCwTone ;
              XGetsCwTone [PotsGate,PotsTalk,PotsHang,CwLink] (Caller,Called)
	   )
           []
	   (
              CwLink  !Called  !GetsCwTone ;
	      XGetsCwTone [PotsGate,PotsTalk,PotsHang,CwLink] (Called,Caller)
           )
        )
     )
     endproc (* CallerTalksToCalled *)

     process PotsConversation [PotsTalk]  (Caller,Called:DecDigit)
     :noexit:=
     (
               PotsTalk  !Caller  !TalksTo  !Called ;
               PotsConversation [PotsTalk] (Caller,Called)
(*
          []
               PotsTalk  !Called  !TalksTo  !Caller ;
               PotsConversation [PotsTalk] (Caller,Called)
*)
     )
     endproc (* PotsConversation *)

     process CalledGetsCwToneFromCaller [PotsGate,PotsTalk,PotsHang,CwLink]
					(Caller,Called : DecDigit) 
     :exit:=
     (
        CwLink  !Called  !GetsCwTone  !Caller ;
        (
           (
              PotsGate  !Caller  !RingsBackFrom  !Called ;
              (
                 (
                    CwLink  !Called  !FlashHook ;
                    (
			  CwLink  ?U:DecDigit  !NotOnLine  !Called ;
                    	  CallerTalksToCalled [PotsGate,PotsTalk,PotsHang,
					       CwLink] (Caller,Called)
		      []  CwLink  ?U:DecDigit  !OnLine  !Called ;
			  XGetsCwTone [PotsGate,PotsTalk,PotsHang,CwLink] 
				      (Called,Caller)
		    )
                 )
                []
                 (
                    CwLink  !Called  !HangsUp  !Caller ;
                    XHangsUpAndYHolds [PotsGate,PotsTalk,PotsHang,CwLink] 
				      (Called,Caller)
                 )
                []
                 (
                    PotsHang  !Caller  !HangsUp ;
                    XTalksAndYHangsUp [PotsGate,PotsHang,CwLink] 
				      (Called,Caller)
                 )
              )
           )
          []
           (
              PotsHang  !Caller  !HangsUp ;
              XTalksAndYHangsUp [PotsGate,PotsHang,CwLink] 
				(Called,Caller)
           )
          []
           (
              CwLink  !Called  !HangsUp  !Caller ;
	      (
		 (
		    PotsHang  !Caller  !HangsUp ;
		    XandYHangUpWithYHolding [PotsGate,PotsHang] (Called,Caller)
		 )
		[]
		 (
		    PotsGate  !Called  !RingsBackFrom  !Called ; 
		    XHangsUpAndYHolds [PotsGate,PotsTalk,PotsHang,CwLink] 
				      (Called,Caller)
		 )
		[]
		 (
		    PotsGate  !Called  !InsteadOf !Called  !RingsFrom  !Caller ;
		    (
			  PotsHang  !Caller  !HangsUp ;
			  XRingsFromYwithYNotOnLine [PotsGate,PotsHang] 
						    (Called,Caller)
		      []  PotsGate  !Caller  !RingsBackFrom  !Called ; 
			  XRingsFromYwithYHolding [PotsGate,PotsTalk,PotsHang,
						   CwLink] (Called,Caller)
		    )
		 )
	      )
           )
        )
     )
     endproc (* CalledGetsCwToneFromCaller *)

     process XGetsCwTone [PotsGate,PotsTalk,PotsHang,CwLink](X,Y:DecDigit) (*C*)
     :exit:=
     (
          PotsConversation [PotsTalk]  (X,Y)
          [>
          (
               (
                    PotsHang  !X  !HangsUp ;
                    UserHangsUpOnPotsHang [PotsHang] (Y)
               )
               []
               (
                    PotsHang  !Y  !HangsUp ;
                    XGetsCwToneAndYHangsUp [PotsHang,CwLink] (X,Y)
               )
               []
               (
                    CwLink  !X  !FlashHook  !Y ;
                    XTalksAndYHolds [PotsGate,PotsTalk,PotsHang,CwLink] (X,Y)
               )
          )
     )
     endproc (* PotsControlXGetsCwTone *)

     process XGetsCwToneAndYHangsUp [PotsHang,CwLink] 
				    (X,Y : DecDigit) 			(* D *)
     :exit:=
     (
               UserHangsUpOnPotsHang [PotsHang] (X)
          []
               CwLink  !X  !FlashHook  !X  !Y ;
               exit 
     )
     endproc (* XGetsCwToneAndYHangsUp *)

     process XTalksAndYHolds [PotsGate,PotsTalk,PotsHang,CwLink]
			     (X,Y:DecDigit)				(* E *)
     :exit:=
     (
        (
           CwLink  !X  !HangsUp  !Y ;
           XHangsUpAndYHolds [PotsGate,PotsTalk,PotsHang,CwLink] (X,Y)
        )
       []
        (
           PotsHang  !Y  !HangsUp ;
           XTalksAndYHangsUp [PotsGate,PotsHang,CwLink] (X,Y)
        )
       []
        (
           CwLink  !X  !FlashHook ;
           (
              (
                 CwLink  ?U:DecDigit  !NotOnLine  !X ;
                 CallerTalksToCalled [PotsGate,PotsTalk,PotsHang,CwLink] (X,Y)
              )
             []
              (
                 CwLink  ?U:DecDigit  !OnLine  !X ;
                 XGetsCwTone [PotsGate,PotsTalk,PotsHang,CwLink] (X,Y)
              )
           )
        )
     )
     endproc (* XTalksAndYHolds *)

     process XHangsUpAndYHolds [PotsGate,PotsTalk,PotsHang,CwLink]
			       (X,Y:DecDigit) 				(* F *)
     :exit:=
     (
        (
           PotsHang  !Y  !HangsUp ;
           XandYHangUpWithYHolding [PotsGate,PotsHang] (X,Y)
        )
       []
        (
           PotsGate  !X  !InsteadOf !X !RingsFrom  !Y ;
	   XRingsFromYwithYHolding [PotsGate,PotsTalk,PotsHang,CwLink] (X,Y)
        )
     )
     endproc (* XHangsUpAndYHolds *)

     process XTalksAndYHangsUp [PotsGate,PotsHang,CwLink] 
			       (X,Y:DecDigit) 				(* G *)
     :exit:=
     (
        CwLink  !X  !FlashHook ;
        (
		CwLink  ?U:DecDigit  !OnLine  !X ;
		XGetsCwToneAndYHangsUp [PotsHang,CwLink] (X,Y)
          []
		CwLink  ?U:DecDigit  !NotOnLine  !X ;
		UserHangsUpOnPotsHang [PotsHang] (X)
        )
       []
        (
           CwLink  !X  !HangsUp  !Y ;
           XandYHangUpWithYHolding [PotsGate,PotsHang] (X,Y) 
        )
     )
     endproc (* XTalksAndYHangsUp *)

     process XandYHangUpWithYHolding [PotsGate,PotsHang] (X,Y:DecDigit)	 (* H *)
     :exit:=
     (
	PotsGate  !X !InsteadOf !X !RingsFrom !Y ;
	XRingsFromYwithYNotOnLine [PotsGate,PotsHang] (X,Y)
     )
     endproc (* XandYHangUpWithYHolding *)

     process XRingsFromYwithYNotOnLine [PotsGate,PotsHang](X,Y:DecDigit) (* I *)
     :exit:=
     (
	   PotsGate  !X  !StopToRingFrom  !Y ;
	   exit
	[]
	   PotsGate  !X  !OffHookToAnswer !Y ;
	   UserHangsUpOnPotsHang [PotsHang] (X)
     )
     endproc (* XRingsFromYwithYNotOnLine *)

     process XRingsFromYwithYHolding [PotsGate,PotsTalk,PotsHang,CwLink] 
				     (X,Y:DecDigit)			 (* N *)
     :exit:=
     (
	(
	   PotsHang !Y !HangsUp ;
	   XRingsFromYwithYNotOnLine [PotsGate,PotsHang] (X,Y)
	)
	[]
	(
	   PotsGate !X !OffHookToAnswer !Y ;
	   CallerTalksToCalled [PotsGate,PotsTalk,PotsHang,CwLink] (Y,X)
	)
     )
     endproc (* XRingsFromYwithYHolding *)


     process NewGlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
     :noexit:=
     (
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                 ( InsertPair (Pair (2,4),{} of SetOfPairs),
		   {} of DecSet ,
                   Insert (2, {} of DecSet) , 
                   {} of SetOfPairs ,
                   {} of SetOfPairs ,
		   {} of SetOfPairs
                 )
        |[CwLink]|

         GlobalCwController [CwLink]
     )
     endproc (* NewGlobalView *)

     process GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
            ( CfblPairs   :SetOfPairs, (* First element forwards to second element *)
              BusyList ,              (* list of busy users               *)
              CwList      : DecSet ,  (* list of users with CallWaiting   *)
              WaitPairs,              (* list of pairs (Caller,Called)    *)
                                      (* where Called is waiting for X	  *)
                                      (* to answer			  *)
              HoldPairs,              (* list of pairs (X,Y) where X made *)
                                      (* a FlashHook and Y is holding     *)
              CwPairs    : SetOfPairs (* list of pairs (T,X) where T is the
                                         'CallWaiting' caller and X is the
                                         'CallWaiting' called             *)
            )
     :noexit:=
     (
             PotsPhases [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            ( CfblPairs,
			      BusyList ,
                              CwList   ,
                              WaitPairs,
                              HoldPairs,
                              CwPairs  
                            )
         []  CwPhases [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            ( CfblPairs,
                              BusyList ,
                              CwList   ,
                              WaitPairs,
                              HoldPairs,
                              CwPairs 
                            )
(*
         []  PrintSets [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            ( CfblPairs,
                              BusyList ,
                              CwList   ,
                              WaitPairs,
                              HoldPairs,
                              CwPairs 
                            )
*)
     )
     where

     process PotsPhases [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                        ( CfblPairs   : SetOfPairs,
                          BusyList ,
                          CwList      : DecSet ,
                          WaitPairs,
                          HoldPairs,
                          CwPairs     : SetOfPairs
                        )
     :noexit:=
     (
        PotsGate  ?Caller : DecDigit  !OffHookToCall
        [     (Caller NotIn BusyList)
          and (SecondElement( Caller , WaitPairs ) eq none)
          and (SecondElement( Caller , HoldPairs ) eq none)
        ] ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
		     Insert (Caller,BusyList) ,
                     CwList   ,
                     InsertPair (Pair (Caller,Caller),WaitPairs),
                     HoldPairs,
                     CwPairs  
                   )
    []  PotsGate  ?Caller : DecDigit  ! GetsTone  ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     WaitPairs,
                     HoldPairs,
                     CwPairs 
                   )
    []  PotsGate  ?Caller : DecDigit  ! Dials  ?Called : DecDigit ;
	HiddenGate  !SecondElement(Called,CfblPairs)  !Called ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     WaitPairs,
                     HoldPairs,
                     CwPairs 
                   )
    []  PotsGate ?Caller:DecDigit !GetsBusySignalFrom ?X:DecDigit !InsteadOf
	?Called:DecDigit
        [   (     (X eq Called)
	      and (SecondElement(X,CfblPairs) eq none)
              and (   (Called IsIn BusyList)
                    or (    (SecondElement(Called,WaitPairs) ne none)
                        and (SecondElement(Called,WaitPairs) ne Caller)
                       )
                    or (     (SecondElement(Called,HoldPairs) ne none)
                         and (SecondElement(Called,HoldPairs) ne none)
                       )
                  )
              and (    (Called NotIn BusyList)
                    or (Called eq Caller)
                    or (Called NotIn CwList)
                    or (FirstElement(Called,WaitPairs) ne none)
                    or (SecondElement(Called,CwPairs) ne none)
                  )
	    )
         or (    (X eq SecondElement(Called,CfblPairs))
             and (    (X IsIn BusyList)
                   or (    (SecondElement(X,WaitPairs) ne none)
                       and (SecondElement(X,WaitPairs) ne Caller)
                      )
                   or (     (SecondElement(X,HoldPairs) ne none)
                        and (SecondElement(X,HoldPairs) ne Caller)
                      )
                 )
	     and (    (Called IsIn BusyList)
                   or (    (SecondElement(Called,WaitPairs) ne none)
                       and (SecondElement(Called,WaitPairs) ne Caller)
                      )
                   or (     (SecondElement(Called,HoldPairs) ne none)
                        and (SecondElement(Called,HoldPairs) ne none)
                      )
                 )
            )
        ] ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     WaitPairs,
                     HoldPairs,
                     CwPairs 
                   )
    []  PotsGate ?X:DecDigit !InsteadOf ?Called:DecDigit !RingsFrom 
	?Caller:DecDigit 
        [    (     (X eq Called)
	       and (X NotIn BusyList)
               and (    (SecondElement(X,WaitPairs) eq none)
                     or (SecondElement(X,WaitPairs) eq Caller)
                   )
               and (    (SecondElement(X,HoldPairs) eq none)
                     or (SecondElement(X,HoldPairs) eq Caller)
                   )
	     )
          or (     (X eq SecondElement(Called,CfblPairs))
               and (     (X NotIn BusyList)
                     and (    (SecondElement(X,WaitPairs) eq none)
                           or (SecondElement(X,WaitPairs) eq Caller)
                         )
                     and (    (SecondElement(X,HoldPairs) eq none)
                           or (SecondElement(X,HoldPairs) eq Caller)
                         )
                   )
               and (    (Called IsIn BusyList)
                     or (    (SecondElement(Called,WaitPairs) ne none)
                         and (SecondElement(Called,WaitPairs) ne Caller)
                        )
                     or (     (SecondElement(Called,HoldPairs) ne none)
                          and (SecondElement(Called,HoldPairs) ne none)
                        )
                   )
             )
        ] ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     InsertPairAtTheEnd( Pair(X,Caller) ,
                     RemovePair ( Pair (Caller,Caller),WaitPairs)),
                     HoldPairs,
                     CwPairs
                   )
    []  PotsGate  ?Caller:DecDigit  !RingsBackFrom  ?Called:DecDigit ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     WaitPairs,
                     HoldPairs,
                     CwPairs 
                   )
    []  PotsGate ?Called:DecDigit !OffHookToAnswer ?Caller:DecDigit ;
        (
           [Caller eq SecondElement(Called,CwPairs)] 
	   ->  GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     Insert (Called,BusyList),
                     CwList   ,
                     RemovePair (Pair (Called,Caller) , WaitPairs ) ,
                     RemovePair (Pair (Called,Caller) , HoldPairs ) ,
                     RemovePair (Pair (Called,Caller) , CwPairs ) 
                   )
	   []
           [Caller ne SecondElement(Called,CwPairs)] 
	   ->  GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     Insert (Called,BusyList),
                     CwList   ,
                     RemovePair (Pair (Called,Caller) , WaitPairs ) ,
                     RemovePair (Pair (Called,Caller) , HoldPairs ) ,
                     CwPairs 
                   )
        )
    []  PotsGate  ?Called:DecDigit  !StopToRingFrom  ?Caller:DecDigit
        [    (Caller NotIn BusyList)
          or (SecondElement(Caller,HoldPairs) eq Called)] ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     RemovePair (Pair (Called,Caller) , WaitPairs ) ,
                     RemovePair (Pair (Called,Caller) , HoldPairs ) ,
                     RemovePair (Pair (Called,Caller) , CwPairs ) 
                   )
    []  PotsTalk  ?Caller : DecDigit  !TalksTo  ?Called : DecDigit ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     WaitPairs,
                     HoldPairs,
                     CwPairs 
                   )
    []  PotsHang  ?X : DecDigit  !HangsUp
        [SecondElement(X,HoldPairs) eq none] ;
        HangUpPhase1 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                    ( X,
                      CfblPairs,
                      Remove (X,BusyList),
                      CwList ,
                      RemovePair (Pair(FirstElement (X,WaitPairs),X),WaitPairs),
                      HoldPairs,
                      CwPairs 
                    )
    )
    endproc (* PotsPhases *)

    process CwPhases [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                     ( CfblPairs   : SetOfPairs,
                       BusyList ,
                       CwList      : DecSet ,
                       WaitPairs,
                       HoldPairs,
                       CwPairs     : SetOfPairs
		     )
    :noexit:=
    (
        CwLink  ?Called:DecDigit !GetsCwTone  ?Caller:DecDigit
        [    (Called IsIn BusyList)
         and (Called ne Caller)
         and (Called IsIn CwList)
         and (FirstElement(Called,WaitPairs) eq none)
         and (SecondElement(Called,CwPairs) eq none)
        ] ;
        CwLink  !Called   !GetsCwTone ;
        GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( CfblPairs,
                     BusyList ,
                     CwList   ,
                     InsertPairAtTheEnd (Pair (Called,Caller),
                     RemovePair (Pair (Caller,Caller),WaitPairs)),
                     InsertPairAtTheEnd (Pair (Called,Caller),HoldPairs),
                     InsertPairAtTheEnd (Pair (Called,Caller),CwPairs) 
                   )
   []   CwLink  ?X:DecDigit  !HangsUp  ?U:DecDigit
        [SecondElement (X,HoldPairs) eq U] ;
        PotsHang  !X   !HangsUp ;
        HangUpPhase1 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                     ( X,
                       CfblPairs,
                       Remove ( X , BusyList ) ,
                       CwList ,
                       RemovePair (Pair(FirstElement(X,WaitPairs),X),WaitPairs),
                       HoldPairs,
                       CwPairs  
                     )
   []   CwLink  ?X:DecDigit  !FlashHook ;
        (
           CwLink  !X  !FlashHook  ?U:DecDigit [U IsIn BusyList];
           CwLink  !U  ?Signal:Signal  !X ;
           FlashHookPhase1 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( X,U,
                     CfblPairs,
                     BusyList ,
                     CwList   ,
                     WaitPairs,
                     HoldPairs,
                     CwPairs 
                   )
        [] CwLink  !X  !FlashHook  !X  ?U:DecDigit [U NotIn BusyList] ;
           CwLink  !U  ?Signal:Signal  !X ;
           FlashHookPhase1 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                   ( X,X,
                     CfblPairs,
                     BusyList ,
                     CwList   ,
                     WaitPairs,
                     HoldPairs,
                     CwPairs 
                   )
        )
     )
     endproc (* CwPhases *)

    process HangUpPhase1 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                         (X           : DecDigit,     (* !X !HangsUp *)
                          CfblPairs   : SetOfPairs,
                          BusyList ,
                          CwList      : DecSet ,
                          WaitPairs,
                          HoldPairs,
                          CwPairs     : SetOfPairs
                         )
    :noexit:=
    (  let T:DecDigit = SecondElement (X,CwPairs) in
       (
          [    (T eq none)
            or (T eq (SecondElement(X,WaitPairs)))
            or (T eq (SecondElement(X,HoldPairs)))
          ]  ->  GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (
                              CfblPairs,
                              BusyList ,
                              CwList   ,
                              WaitPairs,
                              HoldPairs,
                              CwPairs 
                            )
         []
          [     (T ne none)
            and (T ne SecondElement(X,WaitPairs))
            and (T ne SecondElement(X,HoldPairs))
          ]  ->  GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (
                              CfblPairs,
                              BusyList ,
                              CwList   ,
                              WaitPairs,
                              HoldPairs,
                              RemovePair (Pair (X,T) , CwPairs ) 
                            )
       )
    )
    endproc (* HangUpPhase1 *)

    process FlashHookPhase1 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                           (X,                (* !X !FlashHook !U *)
                            U           : DecDigit ,
                            CfblPairs   : SetOfPairs,
                            BusyList ,
                            CwList      : DecSet ,
                            WaitPairs,
                            HoldPairs,
                            CwPairs     : SetOfPairs
                           )
    :noexit:=
    (  let T:DecDigit = SecondElement(X,CwPairs) in
       (
          [    (T eq SecondElement(X,HoldPairs))
            or (     (T IsIn BusyList)
                 and (SecondElement(T,HoldPairs) ne X)
               )
          ] -> FlashHookPhase2 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                ( X,U,
                  CfblPairs,
                  BusyList,
                  CwList  ,
                  RemovePair (Pair (X,SecondElement(X,WaitPairs)),WaitPairs),
                  RemovePair (Pair (X,SecondElement(X,HoldPairs)),HoldPairs),
                  CwPairs  
                )
       [] [     (T ne SecondElement(X,HoldPairs))
            and (    (T NotIn BusyList)
                  or (SecondElement(T,HoldPairs) eq X)
                )
          ] -> FlashHookPhase2 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                ( X,U,
                  CfblPairs,
                  BusyList,
                  CwList  ,
                  RemovePair (Pair (X,SecondElement(X,WaitPairs)),WaitPairs),
                  RemovePair (Pair (X,SecondElement(X,HoldPairs)),HoldPairs),
                  RemovePair (Pair (X,T),CwPairs ) 
                )
       )
    )
    endproc (* FlashHookPhase1*)

    process FlashHookPhase2 [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                           (X,                (* !X !FlashHook !U *)
                            U           : DecDigit ,
                            CfblPairs   : SetOfPairs,
                            BusyList ,
                            CwList      : DecSet ,
                            WaitPairs,
                            HoldPairs,
                            CwPairs     : SetOfPairs
                           )
    :noexit:=
    (
        [    (U NotIn BusyList)
          or (U eq X)
        ] ->  GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                        (
                          CfblPairs,
                          BusyList ,
                          CwList   ,
                          WaitPairs,
                          HoldPairs,
                          CwPairs 
                        )
       []
        [     (U IsIn BusyList)
          and (U ne X)
        ] ->  GlobalView [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                        (
                          CfblPairs,
                          BusyList ,
                          CwList   ,
                          WaitPairs,
                          InsertPairAtTheEnd (Pair (X,U) , HoldPairs ),
                          CwPairs 
                        )
    )
    endproc (* FlashHookPhase2 *)

    endproc (* GlobalView *)

    process GlobalCwController [CwLink]
    :noexit:=
    (
        CwLink  ?Called:DecDigit  !GetsCwTone  ?Caller:DecDigit ;
        CwLink  !Called  !GetsCwTone ;
        (
            NewCwControllerTX [CwLink] (Caller,Called)
            |||
            GlobalCwController [CwLink]
        )
    )
    endproc (* GlobalCwController *)

    process NewCwControllerTX [CwLink] (T,X : DecDigit )
    :noexit:=
    (
       (
          CwLink  !X  !HangsUp  !T ;
          stop
       )
      []
       (
          CwLink  !X  !FlashHook ;
          (
             (
                CwLink  !X  !FlashHook  !X  ?Y:DecDigit [Y ne T] ;
                CwLink  ?U:DecDigit  !NotOnLine  !X;
                stop
             )
            []
             (
                CwLink  !X  !FlashHook  ?Y:DecDigit [Y ne X] ;
                CwLink  ?U:DecDigit  !OnLine  !X ;
                CwControllerTX_TTalksToX [CwLink] (T,X,Y)
             )
          )
       )
    )
    endproc (* NewCwControllerTX *)

    process CwControllerTX_TTalksToX [CwLink] (T,X,Y:DecDigit)
    :noexit:=
    (
       (
          CwLink  !X  !HangsUp  ?U:DecDigit [U ne T] ;
          stop
       )
      []
       (
          CwLink  !X  !FlashHook ;
          (
             (
                CwLink  !X  !FlashHook  !X  !T ;
                CwLink  ?U:DecDigit  !NotOnLine  !X ;
                stop
             )
            []
             (
                CwLink  !X  !FlashHook  !T ;
                CwLink  ?U:DecDigit  !OnLine  !X ;
                NewCwControllerTX [CwLink] (T,X)
             )
          )
       )
    ) endproc (* CwControllerTX_TTalksToX *)

(*
    process PrintSets [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                      ( Cfblpairs : SetOfPairs,
			B, 
                        Cw        : DecSet ,
                        Wpairs,
                        HPairs,
                        Cwpairs   : SetOfPairs
                      )
    :noexit:=
    (   hide CfblPairs, BusyList, CwList, WaitPairs, HoldPairs, CwPairs in
        (
                CfblPairs  !Cfblpairs;
                GlobalView  [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (Cfblpairs,B,Cw,Wpairs,Hpairs,Cwpairs)
              []
                BusyList    !B;
                GlobalView  [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (Cfblpairs,B,Cw,Wpairs,Hpairs,Cwpairs)
              []
                CwList    !Cw;
                GlobalView  [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (Cfblpairs,B,Cw,Wpairs,Hpairs,Cwpairs)
              []
                WaitPairs    !Wpairs;
                GlobalView  [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (Cfblpairs,B,Cw,Wpairs,Hpairs,Cwpairs)
              []
                HoldPairs    !Hpairs;
                GlobalView  [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (Cfblpairs,B,Cw,Wpairs,Hpairs,Cwpairs)
              []
                CwPairs    !Cwpairs;
                GlobalView  [PotsGate,PotsTalk,PotsHang,CwLink,HiddenGate]
                            (Cfblpairs,B,Cw,Wpairs,Hpairs,Cwpairs)
       )
    )
    endproc
*)

endspec
 

